package cz.mendelu.dnaAnalyser.analyse.g4hunter;

import cz.mendelu.dnaAnalyser.analyse.g4hunter.quadruplex.*;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.utils.analyse.BadParameterAnalyseException;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.batch.BatchService;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.utils.repository.BaseRepository;
import cz.mendelu.dnaAnalyser.utils.service.ModelService;
import cz.mendelu.dnaAnalyser.sequence.NucleicBuffer;
import cz.mendelu.dnaAnalyser.sequence.Sequence;

import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.cursor.Cursor;
import org.apache.ibatis.session.RowBounds;
import org.apache.ibatis.session.SqlSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Service;

import java.util.*;

@Service
@Slf4j
public class G4HunterService extends ModelService<G4Hunter> {

    @Autowired
    private G4HunterRepository g4HunterRepository;

    @Autowired
    private SequenceDataRepository sequenceDataRepository;

    @Autowired
    private QuadruplexDataSessionService quadruplexDataSessionService;

    @Autowired
    private BatchService batchService;

    @Autowired
    private G4HunterAnalyser g4HunterAnalyser;

    @Override
    protected BaseRepository<G4Hunter, UUID> getRepository() {
        return g4HunterRepository;
    }

    public G4Hunter executeG4HunterAnalyse(G4Hunter g4Hunter) {
        if (g4Hunter.getSequence().getLength() < g4Hunter.getWindowSize()) {
            throw new BadParameterAnalyseException(String
                    .format("The window (%d) must be smaller than the sequence length (%d).",
                            g4Hunter.getWindowSize(),
                            g4Hunter.getSequence().getLength()));
        }

        g4Hunter = g4HunterRepository.save(g4Hunter);

        batchService.createBatch(g4Hunter.getOwner(), g4Hunter)
                .execute(uuid -> {
                    G4Hunter analyse = g4HunterRepository.findById(uuid)
                            .orElseThrow(() -> new NotFoundException(G4Hunter.class, uuid));

                    Sequence sequence = analyse.getSequence();
                    SequenceData sequenceData = sequenceDataRepository.load(sequence);
                    int window = analyse.getWindowSize();
                    if (sequence.getLength() < window) {
                        throw new IllegalArgumentException(String.format("The window (%d) must be smaller than the sequence length (%d).", window, sequence.getLength()));
                    }

                    float threshold = analyse.getThreshold();
                    try (SqlSession sqlSession = quadruplexDataSessionService.openDataSession(analyse)) {
                        int[] score = analyseQuadruplex(analyse.getId(), sqlSession, sequenceData, window, threshold);
                        sqlSession.commit();

                        long resultCount = aggregateQuadruplex(analyse.getId(), sqlSession, sequenceData, score);
                        sqlSession.commit();

                        analyse.setResultCount(resultCount);
                        analyse.setFrequency(((float) resultCount / (float) sequence.getLength()) * 1000f);
                        analyse.finishNow();
                        sqlSession.commit();
                    }
                    g4HunterRepository.save(analyse);
                }).start();
        return g4Hunter;
    }

    private int[] analyseQuadruplex(UUID uuid, SqlSession sqlSession, SequenceData sequenceData, int window, float threshold) {
        QuadruplexMapper quadruplexMapper = sqlSession.getMapper(QuadruplexMapper.class);
        NucleicBuffer nucleicBuffer = sequenceData.iterator();
        log.debug("G4Hunter Analyse {}: calculate basic score.", uuid);
        int[] baseScore = g4HunterAnalyser.baseScore(nucleicBuffer);
        log.debug("G4Hunter Analyse {}: calculate windows score.", uuid);
        float[] score = g4HunterAnalyser.windowsScore(baseScore, window);
        log.debug("G4Hunter Analyse {}: Store result to db", uuid);
        long resultCounter = 0;
        for (int i = 0; i < score.length; i++) {
            if (Math.abs(score[i]) >= threshold) {
                NucleicBuffer buffer = sequenceData.getBuffer(i, window);
                Quadruplex quadruplex = new Quadruplex(null, i, window, score[i], buffer.toPlain());
                quadruplexMapper.insert(quadruplex);
                resultCounter++;

                if (resultCounter % 1000 == 0) {
                    log.debug("Anlyse Quadruplex {}, found: {}, analyse {}% of sequence", uuid, resultCounter, (int)(((double)i / (double)score.length) * 100));
                    sqlSession.commit();
                }
            }
        }
        return baseScore;
    }

    private long aggregateQuadruplex(UUID uuid, SqlSession sqlSession, SequenceData sequenceData, int[] baseScore) {
        QuadruplexMapper quadruplexMapper = sqlSession.getMapper(QuadruplexMapper.class);
        AggregateQuadruplexMapper aggregateQuadruplexMapper = sqlSession.getMapper(AggregateQuadruplexMapper.class);

        int totalQuadruplex = quadruplexMapper.countAll();
        int currentQuadruplex = 0;

        long resultCounter = 0;
        AggregateQuadruplex aggregateQuadruplex = null;
        int nextPostiont = Integer.MIN_VALUE;
        Cursor<Quadruplex> quadruplexCursor = quadruplexMapper.getCursor();
        Iterator<Quadruplex> quadruplexIterator = quadruplexCursor.iterator();
        while (quadruplexIterator.hasNext()) {
            Quadruplex quadruplex = quadruplexIterator.next();
            currentQuadruplex++;

            if (aggregateQuadruplex != null && nextPostiont < quadruplex.getPosition()) {
                saveAggregateQuadruplex(aggregateQuadruplexMapper, aggregateQuadruplex, nextPostiont, sequenceData, baseScore);
                resultCounter++;

                if (resultCounter % 100 == 0) {
                    log.debug("Aggregate Quadruplex {}, aggregate: {}, analyse {}% of founded Quadruplex", uuid, resultCounter, (int)(((double)currentQuadruplex / (double)totalQuadruplex) * 100));
                    sqlSession.commit();
                }

                aggregateQuadruplex = null;
            }

            if (aggregateQuadruplex == null) {
                aggregateQuadruplex = new AggregateQuadruplex(quadruplex);
                nextPostiont = quadruplex.getPosition() + 1;
                continue;
            }

            aggregateQuadruplex.addSubScore(quadruplex.getScore());
            nextPostiont++;

        }

        if (aggregateQuadruplex != null) {
            saveAggregateQuadruplex(aggregateQuadruplexMapper, aggregateQuadruplex, nextPostiont, sequenceData, baseScore);
            resultCounter++;
        }

        return resultCounter;
    }

    private void saveAggregateQuadruplex(AggregateQuadruplexMapper aggregateQuadruplexMapper, AggregateQuadruplex aggregateQuadruplex, int nextPosition, SequenceData sequenceData, int[] baseScore) {
        int position = aggregateQuadruplex.getPosition();
        int length = nextPosition - aggregateQuadruplex.getPosition() + aggregateQuadruplex.getLength() - 1;
        NucleicBuffer nucleicBuffer = sequenceData.getBuffer(position, length);
        float score = baseScore[position + length - 1] - ((position > 1) ? baseScore[position - 1] : 0);
        score /= length;

        aggregateQuadruplex.setLength(length);
        aggregateQuadruplex.setSequence(nucleicBuffer.toPlain());
        aggregateQuadruplex.setScore(score);
        aggregateQuadruplex.setAbsScore(Math.abs(score));
        aggregateQuadruplexMapper.insert(aggregateQuadruplex);
    }

    public Page<AggregateQuadruplex> analyseQuadruplex(UUID id, Integer sequenceStart, Integer sequenceLength, Pageable pageable) {
        G4Hunter analyse = findOne(id);
        int start = (sequenceStart == null) ? 0 : sequenceStart;
        int end = start + ((sequenceLength == null) ? analyse.getSequence().getLength() : sequenceLength);
        try (SqlSession sqlSession = quadruplexDataSessionService.openDataSession(analyse)) {
            AggregateQuadruplexMapper aggregateQuadruplexMapper = sqlSession.getMapper(AggregateQuadruplexMapper.class);
            RowBounds rowBounds = new RowBounds((int) pageable.getOffset(), pageable.getPageSize());
            String orderName = "POSITION";
            Sort.Direction orderDirection = Sort.Direction.ASC;
            Sort.Order order = getSortOrderOrNull(pageable);
            if (order != null) {
                orderName = order.getProperty().toUpperCase();
                orderDirection = order.getDirection();
            }

            List<AggregateQuadruplex> list = (orderDirection == Sort.Direction.ASC)
                    ? aggregateQuadruplexMapper.getAllBetweenPositionOrderAsc(start, end, orderName, rowBounds)
                    : aggregateQuadruplexMapper.getAllBetweenPositionOrderDesc(start, end, orderName, rowBounds);
            int total = aggregateQuadruplexMapper.countAllBetweenPosition(start, end);
            Page<AggregateQuadruplex> page = new PageImpl<>(list, pageable, total);
            return page;
        }
    }

    private Sort.Order getSortOrderOrNull(Pageable pageable) {
        if (pageable == null) return null;
        if (pageable.getSort() == null) return null;
        if (pageable.getSort().iterator() == null) return null;
        Iterator<Sort.Order> orderIterator = pageable.getSort().iterator();
        if (orderIterator.hasNext() == false) return null;
        return orderIterator.next();
    }

    public Heatmap getHeatmap(UUID id, Integer segments, Integer from, Integer to) {        G4Hunter analyse = findOne(id);
        Heatmap heatmap = quadruplexDataSessionService.heatmap(AggregateQuadruplexMapper.class, analyse, segments, from, to);
        return heatmap;
    }

    public List<Quadruplex> getQuadruplex(G4Hunter analyse){
        List<Quadruplex> list;
        try (SqlSession sqlSession = quadruplexDataSessionService.openDataSession(analyse)) {
            QuadruplexMapper quadruplexMapper = sqlSession.getMapper(QuadruplexMapper.class);
            list = quadruplexMapper.getAllItems();
        }
        return list;
    }

    public List<AggregateQuadruplex> getAggregateQuadruplex(G4Hunter analyse){
        List<AggregateQuadruplex> list;
        try (SqlSession sqlSession = quadruplexDataSessionService.openDataSession(analyse)) {
            AggregateQuadruplexMapper aggregateQuadruplexMapper = sqlSession.getMapper(AggregateQuadruplexMapper.class);
            list = aggregateQuadruplexMapper.getAllItems();
        }
        return list;
    }

    public G4Hunter modifyTags(UUID id, Set<String> tags) {
        G4Hunter analyse = g4HunterRepository
                .findById(id)
                .orElseThrow(() -> new NotFoundException(G4Hunter.class, id));
        analyse.setTags(tags);
        g4HunterRepository.save(analyse);
        return analyse;
    }
}
