package cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap;

import cz.mendelu.dnaAnalyser.utils.analyse.data.DataLocalisedMapper;
import cz.mendelu.dnaAnalyser.utils.analyse.data.Localised;
import cz.mendelu.dnaAnalyser.utils.analyse.model.Analyse;
import cz.mendelu.dnaAnalyser.utils.analyse.service.DataSessionService;
import org.apache.ibatis.session.RowBounds;
import org.apache.ibatis.session.SqlSession;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

@Service
public interface HeatmapService extends DataSessionService {

    default <T extends DataLocalisedMapper> Heatmap heatmap(Class<T> mapperClass, Analyse analyse, Integer requestSegmentCount, Integer from, Integer to) {        assert analyse.getSequence() != null && analyse.getSequence().getLength() > 0 : "Analyse has no sequence or sequence length is 0";
        assert analyse.getFinished() != null : "Analyse not done";
        int sequenceSize = to - from;
        int segmentSize = calculateSegmentSize(sequenceSize, requestSegmentCount);
        int segmentsCount = calculateSegmentsCount(sequenceSize, requestSegmentCount, segmentSize);
        Heatmap heatmap = new Heatmap(segmentsCount, segmentSize);
        List<Segment> segments = heatmap.getData();

        try (SqlSession session = openDataSession(analyse)) {
            DataLocalisedMapper mapper = session.getMapper(mapperClass);
            List<Localised> list = mapper.getAllBetweenPositionOrderAsc(from, to, "POSITION", new RowBounds());

            // Calculate counts
            list.forEach(l -> segments.get(segmentIndex(l.getPosition(), segmentSize, segmentsCount)).incCount());

            // Calculate coverage
            int[] map = new int[sequenceSize];
            list.forEach(localised -> IntStream
                    .range(localised.getPosition(), localised.getPosition() + localised.getLength())
                    .forEach(i -> map[i % sequenceSize]++));
            for (int i = 0; i < map.length; i++) {
                Segment segment = segments.get(segmentIndex(i, segmentSize, segmentsCount));
                if (map[i] > 0) {
                    segment.incCoverage();
                }
            }
            segments.forEach(s -> s.divCoverage(segmentSize));



            // Calculate max
            Segment max = new Segment();
            max.setCount(segments.stream()
                    .mapToDouble(Segment::getCount)
                    .max().getAsDouble());
            max.setCoverage(segments.stream()
                    .mapToDouble(Segment::getCoverage)
                    .max().getAsDouble());
            heatmap.getMetaData().setMax(max);

            // Calculate min
            Segment min = new Segment();
            min.setCount(segments.stream()
                    .mapToDouble(Segment::getCount)
                    .min().getAsDouble());
            min.setCoverage(segments.stream()
                    .mapToDouble(Segment::getCoverage)
                    .min().getAsDouble());
            heatmap.getMetaData().setMin(min);

            // Calculate avg
            Segment avg = new Segment();
            avg.setCount(segments.stream()
                    .mapToDouble(Segment::getCount)
                    .average().getAsDouble());
            avg.setCoverage(segments.stream()
                    .mapToDouble(Segment::getCoverage)
                    .average().getAsDouble());
            heatmap.getMetaData().setAvg(avg);
            return heatmap;
        }
    }

    static int calculateSegmentSize(int sequenceSize, int requestCount) {
        return (sequenceSize % requestCount == 0)
                ? sequenceSize / requestCount
                : (sequenceSize / requestCount) + 1;
    }

    static int calculateSegmentsCount(int sequenceSize, int requestCount, int segmentSize) {
        return sequenceSize / segmentSize;
    }

    static int segmentIndex(int position, int segmentSize, int segmentsCount) {
        return (position / segmentSize) % segmentsCount;
    }

    /*
    public <T> Heatmap calculateHeadMap(Class<T> clazz, Integer data, int sequenceSize) {

        Heatmap.MetaData metaData = new Heatmap.MetaData();
        heatmap.setMetaData(metaData);
        float segmentSize = (float)sequenceSize / (float)data;
        //Query query = entityManager.createQuery("Select r.position / :size, count(r) from " + clazz.getName() + " r group by r.position / :size");
        //query.setParameter("size", segmentSize);
        //List result = query.getResultList();
        int[] map = new int[sequenceSize];
        findAll(clazz).stream()
                .map(o -> (Localised)o)
                .forEach(localised -> IntStream
                        .range(localised.getPosition(), localised.getPosition() + localised.getLength())
                        .forEach(i -> map[i % sequenceSize]++)
                );
        List<Heatmap.Segment> cells = new ArrayList<>(data);
        for (int i = 0; i < data; i++) {
            cells.add(new Heatmap.Segment());
        }
        for (int i = 0; i < sequenceSize; i++) {
            int c = Math.round((float)i / segmentSize) % data;
            Heatmap.Segment cell = cells.get(c);
            if (map[i] > cell.getCount()) {
                cell.setCount(map[i]);
            }
            if (map[i] > 0) {
                cell.setCoverage(cell.getCount() + 1);
            }
        }
        for (int i = 0; i < data; i++) {
            Heatmap.Segment cell = cells.get(i);
            cell.setCoverage(Math.min(cell.getCoverage() / segmentSize, 1d));
        }
        heatmap.setData(cells);
        metaData.setSegmentSize(segmentSize);
        metaData.setSegmentCount(data);

        // Calculate max
        Heatmap.Segment max = new Heatmap.Segment();
        max.setCount(cells.stream()
                .mapToDouble(Heatmap.Segment::getCount)
                .max().getAsDouble());
        max.setCoverage(cells.stream()
                .mapToDouble(Heatmap.Segment::getCoverage)
                .max().getAsDouble());
        metaData.setMax(max);

        // Calculate min
        Heatmap.Segment min = new Heatmap.Segment();
        min.setCount(cells.stream()
                .mapToDouble(Heatmap.Segment::getCount)
                .min().getAsDouble());
        min.setCoverage(cells.stream()
                .mapToDouble(Heatmap.Segment::getCoverage)
                .min().getAsDouble());
        metaData.setMin(min);

        // Calculate avg
        Heatmap.Segment avg = new Heatmap.Segment();
        avg.setCount(cells.stream()
                .mapToDouble(Heatmap.Segment::getCount)
                .average().getAsDouble());
        avg.setCoverage(cells.stream()
                .mapToDouble(Heatmap.Segment::getCoverage)
                .average().getAsDouble());
        metaData.setAvg(avg);
        return heatmap;
    }


     */


}
